import numpy as np
import ctypes
from scipy.sparse import issparse, isspmatrix_csr
import warnings
from sklearn.base import BaseEstimator
from . import _wrapper_double, _wrapper_float


class LinearRegression(BaseEstimator):
    """
    Linear Regression
    
    Typical Linear Regression model, which keeps track of the aggregated data
    needed to obtain the closed-form solution in a way that calling 'partial_fit'
    multiple times would be equivalent to a single call to 'fit' with all the data.

    Note
    ----
    Doing linear regression this way requires both memory and computation time
    which scale quadratically with the number of columns/features/variables. As
    such, the class will by default use C 'float' types (typically ``np.float32``)
    instead of C 'double' (``np.float64``), in order to save memory.

    Parameters
    ----------
    lambda_ : float
        Strenght of the L2 regularization.
    fit_intercept : bool
        Whether to add an intercept term to the formula. If passing 'True', it will
        be the last entry in the coefficients.
    method : str, one of 'chol' or 'sm'
        Method used to fit the model. Options are:

        ``'chol'``:
            Uses the Cholesky decomposition to solve the linear system from the
            least-squares closed-form each time 'fit' or 'partial_fit' is called.
            This is likely to be faster when fitting the model to a large number
            of observations at once, and is able to better exploit multi-threading.
        ``'sm'``:
            Starts with an inverse diagonal matrix and updates it as each
            new observation comes using the Sherman-Morrison formula, thus
            never explicitly solving the linear system, nor needing to calculate
            a matrix inverse. This is likely to be faster when fitting the model
            to small batches of observations. Be aware that with this method, it
            will add regularization to the intercept if passing 'fit_intercept=True'.
    calc_inv : bool
        When using ``method='chol'``, whether to also produce a matrix inverse, which
        is required for using the LinUCB and LinTS prediction modes. Ignored when
        passing ``method='sm'`` (the default).
    use_float : bool
        Whether to use C 'float' type for the required matrices. If passing 'False',
        will use C 'double'. Be aware that memory usage for this model can grow
        very large.
    copy_X : bool
        Whether to make deep copies of the 'X' input passed to the model's methods.
        If passing 'False', the 'X' object might be modified in-place. Note that
        if passing 'False', passing 'X' which is a non-contiguous subset of a
        larger array (e.g. 'X[:10, :100]') might provide the wrong results.

    Attributes
    ----------
    coef_ : array(n) or array(n+1)
        The obtained coefficients. If passing 'fit_intercept=True', the intercept
        will be at the last entry.
    """
    def __init__(self, lambda_=1., fit_intercept=True, method="sm", calc_inv=True,
                 use_float=True, copy_X=True):
        self.lambda_ = lambda_
        self.fit_intercept = fit_intercept
        self._method = method
        self._calc_inv = bool(calc_inv)
        self._use_float = bool(use_float)
        self.copy_X = bool(copy_X)

        self._set_dtype(force_cast=False)
        self.coef_ = np.empty(0, dtype=self._dtype)
        self._XtX = np.empty((0,0), dtype=self._dtype)
        self._invXtX = np.empty((0,0), dtype=self._dtype)
        self._XtY = np.empty(0, dtype=self._dtype)
        self._bufferX = np.empty(0, dtype=self._dtype)
        self._n = 0

        self.is_fitted_ = False

    def _set_dtype(self, force_cast=False):
        self._dtype = ctypes.c_float if self._use_float else ctypes.c_double

        if force_cast:
            if self.coef_.dtype != self._dtype:
                self.coef_ = self.coef_.astype(self._dtype)
            if self._XtX.dtype != self._dtype:
                self._XtX = self._XtX.astype(self._dtype)
            if self._invXtX.dtype != self._dtype:
                self._invXtX = self._invXtX.astype(self._dtype)
            if self._XtY.dtype != self._dtype:
                self._XtY = self._XtY.astype(self._dtype)
            if self._bufferX.dtype != self._dtype:
                self._bufferX = self._bufferX.astype(self._dtype)

    @property
    def use_float(self):
        return self._use_float
    @use_float.setter
    def use_float(self, value):
        self._use_float = use_float
        self._set_dtype(force_cast=True)

    @property
    def method(self):
        return self._method
    
    @method.setter
    def method(self, value):
        assert value in ["chol", "sm"]
        if self._method != value:
            cy_funs = _wrapper_float if self._use_float else _wrapper_double
            if self._method == "sm":
                self._XtX = np.empty(self._invXtX.shape, dtype=self._dtype)
                cy_funs.get_matrix_inv(self._invXtX, self._XtX)
            elif (not self._calc_inv):
                self._invXtX = np.empty(self._XtX.shape, dtype=self._dtype)
                cy_funs.get_matrix_inv(self._XtX, self._invXtX)
                self._calc_inv = True
        self._method = value

    @property
    def calc_inv(self):
        return self._calc_inv
    
    @calc_inv.setter
    def calc_inv(self, value):
        if bool(value) != self._calc_inv:
            if self._method == "chol":
                self._invXtX = np.empty(self._XtX.shape, dtype=self._dtype)
                cy_funs.get_matrix_inv(self._XtX, self._invXtX)
            self._calc_inv = bool(value)


    def _process_X_y_w(self, X, y, sample_weight, only_X=False):
        if X.dtype != self._dtype:
            X = X.astype(self._dtype)
        if not issparse(X):
            Xcsr = None
            if X.__class__.__name__ == "DataFrame":
                X = X.to_numpy()
            if self.copy_X:
                X = X.copy()
            X = np.array(X)
            if len(X.shape) != 2:
                raise ValueError("'X' must be 2-dimensional")
        else:
            if not isspmatrix_csr(X):
                warnings.warn("Sparse matrices only supported in CSR format. Input will be converted.")
                X = csr_matrix(X)
            Xcsr = X.copy()
            X = np.empty((0,0), dtype=self._dtype)
            if len(Xcsr.shape) != 2:
                raise ValueError("'X' must be 2-dimensional")
        if only_X:
            return X, Xcsr

        if issparse(y):
            raise ValueError("Sparse 'y' not supported.")
        if issparse(sample_weight):
            raise ValueError("Sparse 'sample_weight' not supported.")

        y = np.array(y).reshape(-1)
        if sample_weight is None:
            w = np.empty(0, dtype=self._dtype)
        else:
            w = np.array(sample_weight).reshape(-1)

        if self.copy_X:
            y = y.copy()
            w = w.copy()

        if Xcsr is None:
            assert X.shape[0] == y.shape[0]
        else:
            if Xcsr.shape[0] != y.shape[0]:
                raise ValueError("'X' and 'y' must have the same number of rows.")

        if sample_weight is not None:
            if X.shape[0] != w.shape[0]:
                raise ValueError("'sample_weight' must have the same number of rows as 'X' and 'y'.")

        if y.dtype != self._dtype:
            y = y.astype(self._dtype)
        if w.dtype != self._dtype:
            w = w.astype(self._dtype)

        return X, Xcsr, y, w

    def fit(self, X, y, sample_weight=None):
        """
        Fit model to data

        Note
        ----
        Calling 'fit' will reset whatever previous data was there. For fitting
        the model incrementally to new data, use 'partial_fit' instead.

        Parameters
        ----------
        X : array(m,n) or CSR matrix(m, n)
            The covariates.
        y : array-like(m)
            The target variable.
        sample_weight : None or array-like(m)
            Observation weights for each row.

        Returns
        -------
        self

        """
        assert self.method in ["chol", "sm"]

        self._n = X.shape[1]
        X, Xcsr, y, w = self._process_X_y_w(X, y, sample_weight)

        cy_funs = _wrapper_float if self._use_float else _wrapper_double
        if (self.method == "chol") or (X.shape[0] >= X.shape[1]):
            self._XtX, self._invXtX, self._XtY, self.coef_ = \
                cy_funs.fit_model_noinv(
                    X, y, w, Xcsr,
                    add_bias=self.fit_intercept,
                    lam = self.lambda_,
                    calc_inv=self.calc_inv
                )
        else:
            self._invXtX, self._XtY, self.coef_ = \
                cy_funs.fit_model_inv(
                    X, y, w, Xcsr,
                    add_bias=self.fit_intercept,
                    lam = self.lambda_
                )

        if self.method != "chol":
            self._bufferX = np.empty(self._n + int(self.fit_intercept), dtype=self._dtype)

        self.is_fitted_ = True
        return self


    def partial_fit(self, X, y, sample_weight=None, *args, **kwargs):
        """
        Fit model incrementally to new data

        Parameters
        ----------
        X : array(m,n) or CSR matrix(m, n)
            The covariates.
        y : array-like(m)
            The target variable.
        sample_weight : None or array-like(m)
            Observation weights for each row.

        Returns
        -------
        self

        """
        if not self.is_fitted_:
            return self.fit(X, y, sample_weight)

        if X.shape[1] != self._n:
            raise ValueError("Number of columns in X doesn't match with previous.")
        X, Xcsr, y, w = self._process_X_y_w(X, y, sample_weight)

        cy_funs = _wrapper_float if self._use_float else _wrapper_double
        if self.method == "chol":
            if self._invXtX.shape[0] == 0:
                self._invXtX = np.empty((self._XtX.shape[0], self._XtX.shape[1]),
                                        dtype=ctypes.c_float if self._use_float else ctypes.c_double)
            cy_funs.update_running_noinv(
                self._XtX, self._XtY, self._invXtX, self.coef_,
                X, y, w, Xcsr,
                add_bias=self.fit_intercept,
                calc_inv=self.calc_inv
            )
        else:
            cy_funs.update_running_inv(
                self._invXtX,
                self._XtY,
                self.coef_,
                self._bufferX,
                X, y, w, Xcsr,
                add_bias=self.fit_intercept,
            )
        return self

    def predict(self, X):
        """
        Make predictions on new data

        Parameters
        ----------
        X : array(m,n) or CSR matrix(m, n)
            The covariates.

        Returns
        -------
        y_hat : array(m)
            The predicted values given 'X'.
        """
        assert self.is_fitted_

        pred = X.dot(self.coef_[:self._n])
        if self.fit_intercept:
            pred[:] += self.coef_[self._n]
        return pred

    def _multiply_x_A_x(self, X):
        X, Xcsr = self._process_X_y_w(X, None, None, only_X=True)
        cy_funs = _wrapper_float if self._use_float else _wrapper_double
        return cy_funs.x_A_x_batch(X, self._invXtX, Xcsr, self.fit_intercept)

    def predict_ucb(self, X, alpha=1.0, add_unfit_noise=False, random_state=None):
        """
        Make an upper-bound prediction on new data

        Make a prediction on new data with an upper bound given by the LinUCB
        formula (be aware that it's not probabilistic like a regular CI).

        Note
        ----
        If using this method, it's recommended to center the 'X' data passed
        to 'fit' and 'partial_fit'. If not centered, it's recommendable to
        lower the ``alpha`` value.

        Parameters
        ----------
        X : array(m,n) or CSR matrix(m, n)
            The covariates.
        alpha : float > 0
            The multiplier for the width of the bound.
        add_unfit_noise : bool
            When making predictions with an unfit model (in this case they are
            given by empty zero matrices except for the inverse diagonal matrix
            based on the regularization parameter), whether to add a very small
            amount of random noise ~ Uniform(0, 10^-12) to it. This is useful in
            order to break ties at random when using multiple models.
        random_state : None, np.random.Generator, or np.random.RandomState
            A NumPy 'Generator' or 'RandomState' object instance to use for generating
            random numbers. If passing 'None', will use NumPy's random
            module directly (which can be made reproducible through
            ``np.random.seed``). Only used when passing ``add_unfit_noise=True``
            and calling this method on a model that has not been fit to data.

        Returns
        -------
        y_hat : array(m)
            The predicted upper bound on 'y' given 'X' and ``alpha``.

        References
        ----------
        .. [1] Chu, Wei, et al. "Contextual bandits with linear payoff functions."
               Proceedings of the Fourteenth International Conference on Artificial Intelligence and Statistics. 2011.
        """
        assert alpha > 0.
        if isinstance(alpha, int):
            alpha = float(alpha)
        assert isinstance(alpha, float)

        if (self.method == "chol") and (not self.calc_inv):
            raise ValueError("Not available when using 'method=\"chol\"' and 'calc_inv=False'.")

        if not self.is_fitted_:
            pred = alpha * np.sqrt(np.einsum("ij,ij->i", X, X) / self.lambda_)
            if add_unfit_noise:
                if random_state is None:
                    noise = np.random.uniform(low=0., high=1e-12, size=X.shape[0])
                else:
                    noise = random_state.uniform(low=0., high=1e-12, size=X.shape[0])
                pred[:] += noise
            return pred

        pred = self.predict(X)
        ci = self._multiply_x_A_x(X)
        pred[:] += alpha * np.sqrt(ci)
        return pred

    def predict_thompson(self, X, v_sq=1.0, sample_unique=False, random_state=None):
        """
        Make a guess prediction on new data

        Make a prediction on new data with coefficients sampled from their
        estimated distribution.

        Note
        ----
        If using this method, it's recommended to center the 'X' data passed
        to 'fit' and 'partial_fit'. If not centered, it's recommendable to
        lower the ``v_sq`` value.

        Parameters
        ----------
        X : array(m,n) or CSR matrix(m, n)
            The covariates.
        v_sq : float > 0
            The multiplier for the covariance matrix. Larger values lead to
            more variable results.
        sample_unique : bool
            Whether to sample different coefficients each time a prediction is to
            be made. If passing 'False', when calling 'predict', it will sample
            the same coefficients for all the observations in the same call to
            'predict', whereas if passing 'True', will use a different set of
            coefficients for each observations. Passing 'False' leads to an
            approach which is theoretically wrong, but as sampling coefficients
            can be very slow, using 'False' can provide a reasonable speed up
            without much of a performance penalty.
        random_state : None, np.random.Generator, or np.random.RandomState
            A NumPy 'Generator' or 'RandomState' object instance to use for generating
            random numbers. If passing 'None', will use NumPy's random
            module directly (which can be made reproducible through
            ``np.random.seed``).

        Returns
        -------
        y_hat : array(m)
            The predicted guess on 'y' given 'X' and ``v_sq``.

        References
        ----------
        .. [1] Agrawal, Shipra, and Navin Goyal.
               "Thompson sampling for contextual bandits with linear payoffs."
               International Conference on Machine Learning. 2013.
        """
        assert self.is_fitted_
        if X.shape[1] != (self.coef_.shape[0]-int(self.fit_intercept)):
            raise ValueError("'X' has wrong number of columns.")
        assert v_sq > 0.
        if isinstance(v_sq, int):
            v_sq = float(v_sq)
        assert isinstance(v_sq, float)

        if (self.method == "chol") and (not self.calc_inv):
            raise ValueError("Not available when using 'method=\"chol\"' and 'calc_inv=False'.")

        tol = 1e-20
        if np.linalg.det(self._invXtX) >= tol:
            cov = self._invXtX
        else:
            cov = self._invXtX.copy()
            n = cov.shape[1]
            for i in range(10):
                cov[np.arange(n), np.arange(n)] += 1e-1
                if np.linalg.det(cov) >= tol:
                    break

        if sample_unique:
            if random_state is None:
                coef = np.random.multivariate_normal(mean=self.coef_,
                                                     cov=v_sq * cov,
                                                     size=X.shape[0])
            else:
                coef = random_state.multivariate_normal(mean=self.coef_,
                                                        cov=v_sq * cov,
                                                        size=X.shape[0])
            if not issparse(X):
                pred = np.einsum("ij,ij->i", X, coef[:,:X.shape[1]])
            else:
                pred = X.multiply(coef[:,:X.shape[1]]).sum(axis=1)
            if self.fit_intercept:
                pred[:] += coef[:,-1]
        else:
            if random_state is None:
                coef = np.random.multivariate_normal(mean=self.coef_,
                                                     cov=v_sq * cov)
            else:
                coef = random_state.multivariate_normal(mean=self.coef_,
                                                        cov=v_sq * cov)
            pred = X.dot(coef[:X.shape[1]])
            if self.fit_intercept:
                pred[:] += coef[-1]

        return pred



import numpy as np
cimport numpy as np
import ctypes
from libc.string cimport memcpy, memset
from cython.parallel cimport prange
from cython cimport boundscheck, nonecheck, wraparound

ctypedef enum CBLAS_ORDER:
    CblasRowMajor = 101
    CblasColMajor = 102

ctypedef CBLAS_ORDER CBLAS_LAYOUT

ctypedef enum CBLAS_TRANSPOSE:
    CblasNoTrans=111
    CblasTrans=112
    CblasConjTrans=113
    CblasConjNoTrans=114

ctypedef enum CBLAS_UPLO:
    CblasUpper=121
    CblasLower=122

ctypedef enum CBLAS_DIAG:
    CblasNonUnit=131
    CblasUnit=132

ctypedef enum CBLAS_SIDE:
    CblasLeft=141
    CblasRight=142

cdef void cblas_tsyrk(
        CBLAS_ORDER Order, CBLAS_UPLO Uplo, CBLAS_TRANSPOSE Trans,
        int N, int K, realtp alpha, realtp *A, int lda, realtp beta, realtp *C, int ldc
    ) nogil:
    cdef char uplo
    cdef char trans
    if (Order == CblasColMajor):
        if (Uplo == CblasUpper):
            uplo = 'U'
        else:
            uplo = 'L'

        if (Trans == CblasTrans):
            trans = 'T'
        elif (Trans == CblasConjTrans):
            trans = 'C'
        else:
            trans = 'N'

    else:
        if (Uplo == CblasUpper):
            uplo = 'L'
        else:
            uplo = 'U'

        if (Trans == CblasTrans):
            trans = 'N'
        elif (Trans == CblasConjTrans):
            trans = 'N'
        else:
            trans = 'T'

    tsyrk(&uplo, &trans, &N, &K, &alpha, A, &lda, &beta, C, &ldc)

cdef void cblas_tgemm(
        CBLAS_ORDER Order, CBLAS_TRANSPOSE TransA, CBLAS_TRANSPOSE TransB, int M, int N, int K,
        realtp alpha, realtp *A, int lda, realtp *B, int ldb, realtp beta, realtp *C, int ldc
    ) nogil:
    cdef char transA
    cdef char transB

    if (Order == CblasColMajor):
        if (TransA == CblasTrans):
            transA = 'T'
        elif (TransA == CblasConjTrans):
            transA = 'C'
        else:
            transA = 'N'

        if (TransB == CblasTrans):
            transB = 'T'
        elif (TransB == CblasConjTrans):
            transB = 'C'
        else:
            transB = 'N'

        tgemm(&transA, &transB, &M, &N, &K, &alpha, A, &lda, B, &ldb, &beta, C, &ldc)

    else:
        if (TransA == CblasTrans):
            transB = 'T'
        elif (TransA == CblasConjTrans):
            transB = 'C'
        else:
            transB = 'N'

        if (TransB == CblasTrans):
            transA = 'T';
        elif (TransB == CblasConjTrans):
            transA = 'C'
        else:
            transA = 'N'

        tgemm(&transA, &transB, &N, &M, &K, &alpha, B, &ldb, A, &lda, &beta, C, &ldc)

cdef void cblas_tgemv(
        CBLAS_ORDER order,  CBLAS_TRANSPOSE TransA,  int m, int n,
        realtp alpha, realtp  *a, int lda,  realtp  *x, int incx,  realtp beta,  realtp  *y, int incy
    ) nogil:
    cdef char trans
    if (order == CblasColMajor):
        if (TransA == CblasNoTrans):
            trans = 'N';
        elif (TransA == CblasTrans):
            trans = 'T'
        else:
            trans = 'C'
        tgemv(&trans, &m, &n, &alpha, a, &lda, x, &incx, &beta, y, &incy)

    else:
        if (TransA == CblasNoTrans):
            trans = 'T'
        elif (TransA == CblasTrans):
            trans = 'N'
        else:
            trans = 'N'

        tgemv(&trans, &n, &m, &alpha, a, &lda, x, &incx, &beta, y, &incy)

cdef void cblas_tsymv(
        CBLAS_LAYOUT layout,
        CBLAS_UPLO Uplo, int N,
        realtp alpha, realtp  *A, int lda,
        realtp  *X, int incX, realtp beta,
        realtp  *Y, int incY
    ) nogil:
    cdef char ul
    if (layout == CblasColMajor):
        if (Uplo == CblasUpper):
            ul = 'U'
        else:
            ul = 'L'
    else:
        if (Uplo == CblasUpper):
            ul = 'L'
        else:
            ul = 'U'
    tsymv(&ul, &N, &alpha, A, &lda, X, &incX, &beta, Y, &incY)

cdef void cblas_tsymm(
        CBLAS_LAYOUT layout, CBLAS_SIDE Side,
        CBLAS_UPLO Uplo, int M, int N,
        realtp alpha, realtp  *A, int lda,
        realtp  *B, int ldb, realtp beta,
        realtp  *C, int ldc
    ) nogil:
    cdef char sd
    cdef char ul

    if (layout == CblasColMajor):
        if (Side == CblasRight):
            sd = 'R'
        else:
            sd = 'L'

        if (Uplo == CblasUpper):
            ul = 'U'
        else:
            ul = 'L'

        tsymm(&sd, &ul, &M, &N, &alpha, A, &lda, B, &ldb, &beta, C, &ldc)

    else:
        if (Side == CblasRight):
            sd = 'L'
        else:
            sd = 'R'

        if (Uplo == CblasUpper):
            ul = 'L'
        else:
            ul = 'U'

        tsymm(&sd, &ul, &N, &M, &alpha, A, &lda, B, &ldb, &beta, C, &ldc)

cdef void cblas_tsyr(
        CBLAS_LAYOUT layout, CBLAS_UPLO Uplo,
        int N, realtp  alpha, realtp  *X,
        int incX, realtp  *A, int lda
    ) nogil:
    cdef char ul

    if (layout == CblasColMajor):
        if (Uplo == CblasLower):
            ul = 'L'
        else:
            ul = 'U'
        tsyr(&ul, &N, &alpha, X, &incX, A, &lda)

    else:
        if (Uplo == CblasLower):
            ul = 'U'
        else:
            ul = 'L'
        tsyr(&ul, &N, &alpha, X, &incX, A, &lda)


cdef void tgemv_dense_sp(
        realtp A[], int n_plusb,
        realtp data[], long indptr[], long indices[], long row,
        realtp outp[], bint add_bias
    ) nogil:
    cdef long i
    cdef long col
    cdef long rowA
    memset(outp, 0, n_plusb*sizeof(realtp))
    for rowA in range(n_plusb):
        for i in range(indptr[row], indptr[row+1]):
            col = indices[i]
            outp[rowA] += A[(col + rowA*n_plusb) if (col >= rowA) else (rowA + col*n_plusb)] * data[i]
    
    cdef int one = 1
    cdef realtp one_realtp = 1.
    if add_bias:
        taxpy(&n_plusb, &one_realtp, A + (n_plusb-1), &n_plusb, outp, &one)

cdef realtp tdot_dense_sp(
        realtp *dense_vec, long row,
        realtp data[], long indptr[], long indices[]
    ) nogil:
    cdef realtp res = 0
    cdef long i
    for i in range(indptr[row], indptr[row+1]):
        res += dense_vec[indices[i]] * data[i]
    return res

cdef realtp* get_ptr_realtp(np.ndarray[realtp, ndim=1] arr):
    return &arr[0]

cdef long* get_ptr_long(np.ndarray[long, ndim=1] arr):
    return &arr[0]

def cast_csr(Xcsr):
    if Xcsr.data.dtype != C_realtp:
        Xcsr.data = Xcsr.data.astype(C_realtp)
    if Xcsr.indptr.dtype != ctypes.c_long:
        Xcsr.indptr = Xcsr.indptr.astype(ctypes.c_long)
    if Xcsr.indices.dtype != ctypes.c_long:
        Xcsr.indices = Xcsr.indices.astype(ctypes.c_long)

def fit_model_noinv(
        np.ndarray[realtp, ndim=2] X,
        np.ndarray[realtp, ndim=1] y,
        np.ndarray[realtp, ndim=1] w,
        Xcsr = None,
        bint add_bias=1,
        realtp lam = 1.,
        bint calc_inv=0
    ):
    cdef int n = X.shape[1] if Xcsr is None else Xcsr.shape[1]
    cdef int n_plusb = n + <int>add_bias
    cdef int m = X.shape[0] if Xcsr is None else Xcsr.shape[0]
    cdef np.ndarray[realtp, ndim=2] XtX = np.empty((n_plusb,n_plusb), dtype=C_realtp)
    cdef np.ndarray[realtp, ndim=1] XtY = np.empty(n_plusb, dtype=C_realtp)

    update_matrices_noinv(
        X, y, w,
        XtX, XtY,
        Xcsr,
        add_bias=add_bias,
        overwrite=1
    )

    if lam != 0.:
        XtX[np.arange(n), np.arange(n)] += lam

    cdef np.ndarray[realtp, ndim=2] XtX_copy
    cdef np.ndarray[realtp, ndim=1] XtY_copy = XtY.copy()

    cdef char lo = 'L'
    cdef int one_int = 1
    cdef int ignore = 0

    if not calc_inv:
        XtX_copy = XtX.copy()
        tposv(&lo, &n_plusb, &one_int, &XtX[0,0], &n_plusb, &XtY[0], &n_plusb, &ignore)
        return XtX_copy, XtX, XtY_copy, XtY
        ## XtX, invXtX, XtY, coef

    else:
        tposv(&lo, &n_plusb, &one_int, &XtX[0,0], &n_plusb, &XtY[0], &n_plusb, &ignore)
        tpotri(&lo, &n_plusb, &XtX[0,0], &n_plusb, &ignore)
        return np.empty((0,0), dtype=C_realtp), XtX, XtY_copy, XtY
        ## dummy, invXtX, XtY, coef


def fit_model_inv(
        np.ndarray[realtp, ndim=2] X,
        np.ndarray[realtp, ndim=1] y,
        np.ndarray[realtp, ndim=1] w,
        Xcsr = None,
        bint add_bias=1,
        realtp lam = 1.
    ):

    cdef int n = X.shape[1] if Xcsr is None else Xcsr.shape[1]
    cdef int n_plusb = n + <int>add_bias

    cdef np.ndarray[realtp, ndim=2] invXtX = np.zeros((n_plusb,n_plusb), dtype=C_realtp)
    cdef np.ndarray[realtp, ndim=1] XtY = np.empty(n_plusb, dtype=C_realtp)

    invXtX[np.arange(n_plusb), np.arange(n_plusb)] = 1. / lam

    if np.isfortran(X):
        X = np.ascontiguousarray(X)
    if X.dtype != C_realtp:
        X = X.astype(C_realtp)

    update_matrices_inv(
        invXtX,
        XtY,
        XtY,
        X, y, w,
        Xcsr = Xcsr,
        add_bias=add_bias,
        overwrite=1
    )

    cdef realtp *ptr_XtY = &XtY[0]
    cdef realtp *ptr_invXtX = &invXtX[0,0]

    cdef np.ndarray[realtp, ndim=1] coefs = np.empty(n_plusb, dtype=C_realtp)
    cblas_tsymv(CblasRowMajor, CblasUpper, n_plusb,
                1., ptr_invXtX, n_plusb, ptr_XtY, 1,
                0., &coefs[0], 1)

    return invXtX, XtY, coefs


def update_running_noinv(
        np.ndarray[realtp, ndim=2] XtX,
        np.ndarray[realtp, ndim=1] XtY,
        np.ndarray[realtp, ndim=2] invXtX,
        np.ndarray[realtp, ndim=1] coef,
        np.ndarray[realtp, ndim=2] X,
        np.ndarray[realtp, ndim=1] y,
        np.ndarray[realtp, ndim=1] w,
        Xcsr = None,
        bint add_bias=1,
        bint calc_inv=0
    ):

    cdef int n = X.shape[1] if Xcsr is None else Xcsr.shape[1]
    cdef int n_plusb = n + <int>add_bias
    cdef int m = X.shape[0] if Xcsr is None else Xcsr.shape[0]

    cdef np.ndarray[realtp, ndim=2] Xw

    cdef char lo = 'L'
    cdef int one_int = 1
    cdef int ignore = 0

    update_matrices_noinv(
        X, y, w,
        XtX, XtY,
        Xcsr = Xcsr,
        add_bias=add_bias,
        overwrite=0
    )

    coef[:] = XtY
    invXtX[:,:] = XtX[:,:]
    tposv(&lo, &n_plusb, &one_int, &invXtX[0,0], &n_plusb, &coef[0], &n_plusb, &ignore)
    if calc_inv:
        tpotri(&lo, &n_plusb, &invXtX[0,0], &n_plusb, &ignore)

def update_running_inv(
        np.ndarray[realtp, ndim=2] invXtX,
        np.ndarray[realtp, ndim=1] XtY,
        np.ndarray[realtp, ndim=1] coefs,
        np.ndarray[realtp, ndim=1] x_vec, ### buffer
        np.ndarray[realtp, ndim=2] X,
        np.ndarray[realtp, ndim=1] y,
        np.ndarray[realtp, ndim=1] w,
        Xcsr = None,
        bint add_bias=1
    ):

    if np.isfortran(X):
        X = np.ascontiguousarray(X)

    cdef int n = X.shape[1] if Xcsr is None else Xcsr.shape[1]
    cdef int n_plusb = n + <int>add_bias
    
    update_matrices_inv(
        invXtX, XtY, x_vec,
        X, y, w,
        Xcsr = Xcsr,
        add_bias=add_bias,
        overwrite=0
    )

    cdef realtp *ptr_XtY = &XtY[0]
    cdef realtp *ptr_invXtX = &invXtX[0,0]
    cblas_tsymv(CblasRowMajor, CblasUpper, n_plusb,
                1., ptr_invXtX, n_plusb, ptr_XtY, 1,
                0., &coefs[0], 1)


def update_matrices_noinv(
        np.ndarray[realtp, ndim=2] X,
        np.ndarray[realtp, ndim=1] y,
        np.ndarray[realtp, ndim=1] w,
        np.ndarray[realtp, ndim=2] XtX,
        np.ndarray[realtp, ndim=1] XtY,
        Xcsr = None,
        bint add_bias=1,
        bint overwrite=0
    ):
    cdef int n = X.shape[1] if Xcsr is None else Xcsr.shape[1]
    cdef int n_plusb = n + <int>add_bias
    cdef int m = X.shape[0] if Xcsr is None else Xcsr.shape[0]

    cdef np.ndarray[realtp, ndim=1] Xsum
    cdef np.ndarray[realtp, ndim=2] Xw

    cdef CBLAS_ORDER x_ord = CblasRowMajor if not np.isfortran(X) else CblasColMajor
    if Xcsr is None:
        if w.shape[0] == 0:
            cblas_tsyrk(x_ord, CblasUpper if (x_ord==CblasRowMajor) else CblasLower, CblasTrans,
                        n, m,
                        1., &X[0,0], n if (x_ord==CblasRowMajor) else m,
                        0. if overwrite else 1., &XtX[0,0], n_plusb)
        else:
            Xw = X * w.reshape((-1,1))
            if np.isfortran(Xw) != np.isfortran(X):
                X = np.ascontiguousarray(X)
                Xw = np.ascontiguousarray(Xw)
                x_ord = CblasRowMajor
            cblas_tgemm(x_ord, CblasTrans, CblasNoTrans,
                        n, n, m,
                        1., &Xw[0,0], n if (x_ord==CblasRowMajor) else m,
                        &X[0,0], n if (x_ord==CblasRowMajor) else m,
                        0. if overwrite else 1., &XtX[0,0], n_plusb)
    else:
        if w.shape[0] == 0:
            if overwrite:
                XtX[:n,:n] = np.array(Xcsr.T.dot(Xcsr).todense())
            else:
                XtX[:n,:n] += np.array(Xcsr.T.dot(Xcsr).todense())
        else:
            Xw_sp = Xcsr.multiply(w.reshape((-1,1)))
            if overwrite:
                XtX[:n,:n] = np.array(Xw_sp.T.dot(Xcsr).todense())
            else:
                XtX[:n,:n] += np.array(Xw_sp.T.dot(Xcsr).todense())

    cdef int one_int = 1
    cdef realtp one = 1.
    if add_bias:
        if overwrite:
            XtX[n,n] = 0

        if w.shape[0] == 0:
            Xsum = X.sum(axis=0)
            XtX[n,n] += <realtp>m
        else:
            Xsum = np.einsum("ij,i->j", X, w)
            XtX[n,n] += <realtp> (w.sum())
        if overwrite:
            tcopy(&n, &Xsum[0], &one_int, &XtX[0,n], &n_plusb)
        else:
            taxpy(&n, &one, &Xsum[0], &one_int, &XtX[0,n], &n_plusb)

    if XtY.shape[0]:
        if w.shape[0] == 0:
            update_XtY(
                XtY,
                X, y, np.empty(0, dtype=C_realtp),
                Xcsr = Xcsr,
                add_bias=add_bias,
                overwrite=overwrite
            )
        else:
            update_XtY(
                XtY,
                Xw if Xcsr is None else X, y, np.empty(0, dtype=C_realtp),
                Xcsr = Xw_sp if Xcsr is not None else None,
                add_bias=add_bias,
                overwrite=overwrite
            )

def update_matrices_inv(
        np.ndarray[realtp, ndim=2] invXtX,
        np.ndarray[realtp, ndim=1] XtY,
        np.ndarray[realtp, ndim=1] x_vec, ###buffer
        np.ndarray[realtp, ndim=2] X,
        np.ndarray[realtp, ndim=1] y,
        np.ndarray[realtp, ndim=1] w,
        Xcsr = None,
        bint add_bias=1,
        bint overwrite=0
    ):

    cdef int n = X.shape[1] if Xcsr is None else Xcsr.shape[1]
    cdef int n_plusb = n + <int>add_bias
    cdef int m = X.shape[0] if Xcsr is None else Xcsr.shape[0]

    if np.isfortran(X):
        X = np.ascontiguousarray(X)

    cdef realtp *prt_X = &X[0,0] if Xcsr is None else NULL
    cdef realtp *ptr_x_vec = &x_vec[0]
    cdef realtp *ptr_XtX_inv = &invXtX[0,0]
    cdef realtp *ptr_w = NULL
    cdef realtp coef
    cdef int one = 1
    cdef realtp one_realtp = 1.

    if w.shape[0]:
        ptr_w = &w[0]

    cdef realtp *ptr_csr_data = NULL
    cdef long *ptr_csr_indptr = NULL
    cdef long *ptr_csr_indices = NULL

    cdef long i
    if Xcsr is None:
        with nogil, boundscheck(False), nonecheck(False), wraparound(False):
            for i in range(m):
                cblas_tsymv(CblasRowMajor, CblasUpper, n,
                            1., ptr_XtX_inv, n_plusb, prt_X + i*n, 1,
                            0., ptr_x_vec, 1)
                if (add_bias):
                    ptr_x_vec[n] = tdot(&n, ptr_XtX_inv + n, &n_plusb, prt_X + i*n, &one)
                    ptr_x_vec[n] += ptr_XtX_inv[n_plusb*n_plusb - 1]
                    taxpy(&n, &one_realtp, ptr_XtX_inv + n, &n_plusb, ptr_x_vec, &one)

                coef = tdot(&n, ptr_x_vec, &one, prt_X + i*n, &one)
                if add_bias:
                    coef += ptr_x_vec[n]
                coef = -1. / (1. + coef)
                if ptr_w != NULL:
                    coef *= ptr_w[i]

                cblas_tsyr(CblasRowMajor, CblasUpper,
                           n_plusb, coef, ptr_x_vec, 1,
                           ptr_XtX_inv, n_plusb)

    else:
        cast_csr(Xcsr)
        ptr_csr_data = get_ptr_realtp(Xcsr.data)
        ptr_csr_indices = get_ptr_long(Xcsr.indices)
        ptr_csr_indptr = get_ptr_long(Xcsr.indptr)
        with nogil, boundscheck(False), nonecheck(False), wraparound(False):
            for i in range(m):
                tgemv_dense_sp(
                    ptr_XtX_inv, n_plusb,
                    ptr_csr_data, ptr_csr_indptr, ptr_csr_indices, i,
                    ptr_x_vec, add_bias
                )
                coef = tdot_dense_sp(ptr_x_vec, i,
                                     ptr_csr_data, ptr_csr_indptr, ptr_csr_indices)
                if add_bias:
                    coef += ptr_x_vec[n]
                coef = -1. / (1. + coef)
                if ptr_w != NULL:
                    coef *= ptr_w[i]
                cblas_tsyr(CblasRowMajor, CblasUpper,
                           n_plusb, coef, ptr_x_vec, 1,
                           ptr_XtX_inv, n_plusb)

    if ptr_w != NULL:
        y = y * w

    if XtY.shape[0]:
        update_XtY(
            XtY,
            X, y, w,
            Xcsr = Xcsr,
            add_bias=add_bias,
            overwrite=overwrite
        )


def update_XtY(
        np.ndarray[realtp, ndim=1] XtY,
        np.ndarray[realtp, ndim=2] X,
        np.ndarray[realtp, ndim=1] y,
        np.ndarray[realtp, ndim=1] w,
        Xcsr = None,
        bint add_bias=1,
        bint overwrite=0
    ):

    cdef CBLAS_ORDER x_ord = CblasRowMajor if not np.isfortran(X) else CblasColMajor
    
    cdef int n = X.shape[1] if Xcsr is None else Xcsr.shape[1]
    cdef int m = X.shape[0] if Xcsr is None else Xcsr.shape[0]

    cdef np.ndarray[realtp, ndim=2] Xw

    if Xcsr is None:
        if w.shape[0] == 0:
            cblas_tgemv(x_ord, CblasTrans,
                        m, n,
                        1., &X[0,0], n if (x_ord==CblasRowMajor) else m,
                        &y[0], 1,
                        0. if overwrite else 1., &XtY[0], 1)
        else:
            Xw = X * w.reshape((-1,1))
            if np.isfortran(Xw) != np.isfortran(X):
                X = np.ascontiguousarray(X)
                Xw = np.ascontiguousarray(Xw)
                x_ord = CblasRowMajor
            cblas_tgemv(x_ord, CblasTrans,
                        m, n,
                        1., &Xw[0,0], n if (x_ord==CblasRowMajor) else m,
                        &y[0], 1,
                        0. if overwrite else 1., &XtY[0], 1)
    
    else:
        if w.shape[0] == 0:
            if overwrite:
                XtY[:n] = Xcsr.T.dot(y)
            else:
                XtY[:n] += Xcsr.T.dot(y)
        else:
            Xw_sp = Xcsr.multiply(w.reshape((-1,1)))
            if overwrite:
                XtY[:n] = Xw_sp.T.dot(y)
            else:
                XtY[:n] += Xw_sp.T.dot(y)

    cdef int one_int = 1
    cdef realtp one = 1.
    if add_bias:
        if overwrite:
            XtY[n] = 0

        if w.shape[0] == 0:
            XtY[n] += y.sum()
        else:
            XtY[n] += tdot(&m, &y[0], &one_int, &w[0], &one_int)
 
def x_A_x_batch(
        np.ndarray[realtp, ndim=2] X,
        np.ndarray[realtp, ndim=2] invXtX,
        Xcsr = None,
        bint add_bias=1,
        bint copy_X=0
    ):
    
    if (X.shape[0] == 0) and (Xcsr.shape[0] == 0):
        return None
    
    cdef int n = X.shape[1] if Xcsr is None else Xcsr.shape[1]
    cdef int n_plusb = n + <int>add_bias
    cdef int m = X.shape[0] if Xcsr is None else Xcsr.shape[0]

    cdef np.ndarray[realtp, ndim=2] tempX
    cdef np.ndarray[realtp, ndim=1] outp = np.empty(m, dtype=C_realtp)

    cdef long i, j
    cdef realtp one = 1.
    cdef int one_int = 1

    cdef realtp *ptr_tempX = NULL
    cdef realtp *ptr_X = NULL
    cdef realtp *ptr_invXtX = &invXtX[0,0]
    cdef realtp *ptr_outp = &outp[0]

    if Xcsr is None:
        X = np.require(X, dtype=C_realtp, requirements=["C_CONTIGUOUS", "OWNDATA", "WRITEABLE"])

        tempX = np.zeros((m, n_plusb), dtype=C_realtp)
        ptr_tempX = &tempX[0,0]
        ptr_X = &X[0,0]

        with nogil, boundscheck(False), nonecheck(False), wraparound(False):
            cblas_tsymm(CblasRowMajor, CblasRight, CblasUpper,
                        m, n,
                        1., ptr_invXtX, n_plusb,
                        ptr_X, n,
                        0., ptr_tempX, n_plusb)
            for i in range(m):
                ptr_outp[i] = tdot(&n, ptr_tempX + i*n_plusb, &one_int, ptr_X + i*n, &one_int)

            if add_bias:
                cblas_tgemv(CblasRowMajor, CblasNoTrans,
                            m, n,
                            2., ptr_X, n, ptr_invXtX + n, n_plusb,
                            1., ptr_outp, 1)
                for i in range(m):
                    ptr_outp[i] += ptr_invXtX[n_plusb*n_plusb - 1]

    else:
        if Xcsr.dtype != C_realtp:
            Xcsr = Xcsr.astype(C_realtp)

        with nogil, boundscheck(False), nonecheck(False), wraparound(False):
            for i in range(n_plusb):
                for j in range(i):
                    ptr_invXtX[j + i*n_plusb] = ptr_invXtX[i + j*n_plusb]

        tempXcsr = Xcsr.dot(invXtX[:n,:n])
        outp[:] = Xcsr.multiply(tempXcsr).sum(axis=1).reshape(-1)

        if add_bias:
            outp[:] += 2. * Xcsr.dot(invXtX[n,:n])
            outp[:] += invXtX[n,n]

    return outp

def get_matrix_inv(
        np.ndarray[realtp, ndim=2] X,
        np.ndarray[realtp, ndim=2] invX
    ):
    if X.shape[0] == 0:
        return None
    invX[:, :] = X[:, :]
    cdef realtp *ptr_X = &X[0,0]
    cdef realtp *ptr_invX = &invX[0,0]
    cdef int n = X.shape[0]
    cdef int ignore = 0
    cdef char lo = 'L'
    with nogil:
        tpotrf(&lo, &n, ptr_invX, &n, &ignore)
        tpotri(&lo, &n, ptr_invX, &n, &ignore)



import ctypes
from scipy.linalg.cython_blas cimport (
    ssyrk as tsyrk,
    sgemm as tgemm,
    sgemv as tgemv,
    ssymv as tsymv,
    ssymm as tsymm,
    ssyr as tsyr,
    scopy as tcopy,
    sdot as tdot,
    saxpy as taxpy,
    sscal as tscal,
    )
from scipy.linalg.cython_lapack cimport (
    sposv as tposv,
    spotri as tpotri,
    spotrf as tpotrf
    )

ctypedef float realtp
C_realtp = ctypes.c_float


include "linreg_untyped.pxi"
